searchNodes=[{"doc":"Implementation of the &lt;~ extructure operator.","ref":"Extructure.html","title":"Extructure","type":"module"},{"doc":"Destructures the right hand side expression into and according to the left side expression. Supports destructure-like implicit keys (with the same name as the variable) as well as optional variables, flexible keyword and key-pair tuple size and order of elements, implicit transformations between a map, list and a key pair tuple. Also supports toggling between the loose mode and the standard Elixir pattern matching (&quot;rigid&quot;) mode where none of the flexibilities are allowed. Fully enforces pattern matching between the left and the right side once taken into account the optional variables and structural transformation. Features Optional variables - Prefix the variable name with an underscore and/or declare it as a function with zero arguments. If used, the underscore is trimmed from the variable's name e.g. an _a is translated to a . Optional variables - Declare the variable as a function taking the default value as its single argument. In the loose (default) mode use maps, keywords and key-pair tuples interchangeably as you see fit. Toggle the loose mode into the rigid mode by using the unary ^ operator left to a map, list or tuple that requires Elixir-like pattern matching. If requiring a loose matching again at a nested level in the structure, use the ^ operator again to toggle back to loose and so on. To match typical tuples or non-keyword lists, switch to the rigid mode with the ^ operator, if only for the tuple in question. For nesting, use the keys as you would in plain Elixir matching: %{ a : %{ b } } &lt;~ [ a : %{ b : 2 } ] # a variable is not set # b variable is set to 2 or %{ a : a = %{ b } } &lt;~ [ a : %{ b : 2 } ] # both a and b variables are set Note Variables with specified keys must be placed (in a keyword list) trailing the ones without the explicit keys: %{ a , b , c : c , d : _d } &lt;~ %{ a : 1 , b : 2 , c : 3 } or [ a , b , c : c , d : _d ] &lt;~ %{ a : 1 , b : 2 , c : 3 } Any errors in the left-side expression are detected in compile time. Usage Instead of: %{ first_name : fist_name , last_name : last_name , } = socket . assigns age = socket . assigns [ :age ] simply use: %{ first_name , last_name , _age } &lt;~ socket . assigns or { first_name , last_name , _age ( 25 ) } &lt;~ socket . assigns or [ first_name , last_name , age ( 25 ) ] &lt;~ socket . assigns See the README.md and extructure_test.exs files for more examples.","ref":"Extructure.html#%3C~/2","title":"Extructure.<~/2","type":"macro"},{"doc":"","ref":"Extructure.html#is_key/1","title":"Extructure.is_key/1","type":"macro"},{"doc":"","ref":"Extructure.html#is_mode/1","title":"Extructure.is_mode/1","type":"macro"},{"doc":"A subset of Elixir Keyword functions slightly modified to support both atom and string keys.","ref":"Extructure.Keylist.html","title":"Extructure.Keylist","type":"module"},{"doc":"","ref":"Extructure.Keylist.html#delete/2","title":"Extructure.Keylist.delete/2","type":"function"},{"doc":"","ref":"Extructure.Keylist.html#keylist?/1","title":"Extructure.Keylist.keylist?/1","type":"function"},{"doc":"","ref":"Extructure.Keylist.html#merge/3","title":"Extructure.Keylist.merge/3","type":"function"},{"doc":"","ref":"Extructure.Keylist.html#reject/2","title":"Extructure.Keylist.reject/2","type":"function"},{"doc":"","ref":"Extructure.Keylist.html#t:key/0","title":"Extructure.Keylist.key/0","type":"type"},{"doc":"","ref":"Extructure.Keylist.html#t:t/0","title":"Extructure.Keylist.t/0","type":"type"},{"doc":"","ref":"Extructure.Keylist.html#t:value/0","title":"Extructure.Keylist.value/0","type":"type"},{"doc":"Extructure is a flexible destructure library for Elixir. By default the library is using loose (flexible) matching, allowing for implicit structural conversions (maps, lists and tuples, from one to another). The key-pair element order in a tuple or a list is also taken loosely by default. Toggling from loose to Elixir-default (&quot;rigid&quot;) mode is done via the ^ operator. Optional variables are also supported with or without a default value.","ref":"readme.html","title":"Extructure","type":"extras"},{"doc":"def deps do [ { :extructure , &quot;~&gt; 0.3.0&quot; } ] end Import the Extructure module in every module where you use it: defmodule Foo do # uses, aliases, and imports import Extructure # .. end","ref":"readme.html#installation","title":"Extructure - Installation","type":"extras"},{"doc":"The docs can be found at HexDocs .","ref":"readme.html#docs","title":"Extructure - Docs","type":"extras"},{"doc":"Fetching two mandatory variables and one optional from the LiveView assigns Assuming a map of socket assigns, a standard pattern matching followed by retrieving an optional variable is shown below: %{ first_name : fist_name , last_name : last_name , } = socket . assigns age = socket . assigns [ :age ] is a one-liner in Extructure: %{ first_name , last_name , _age } &lt;~ socket . assigns Implicit transformation between maps, lists and tuples with key pairs Given the Extructure's loose treatment of structures in terms of their interchangeability, the former can be expressed in a more readable manner: { first_name , last_name , _age } &lt;~ socket . assigns or [ first_name , last_name , _age ] &lt;~ socket . assigns Default values An optional variable can be written as a function taking zero or one arguments, with a single argument being the default value, and/or as the variable name prefixed with a single underscore character _ . [ first_name , last_name , age ( 25 ) ] &lt;~ socket . assigns or [ first_name , last_name , _age ( 25 ) ] &lt;~ socket . assigns Flexible keyword list or tuple size and element order [ b , a ] &lt;~ [ a : 1 , b : 2 , c : 3 ] # =&gt; [ b: 2, a: 1] { b , a } &lt;~ { { :a , 1 } , { :b , 2 } , { :c , 3 } } # =&gt; { { :b, 2}, { :a, 1}} Flexible keyword list head | tail extraction [ b | rest ] &lt;~ [ a : 1 , b : 2 , c : 3 ] # =&gt; [ b: 2, a: 1, c: 3] [ b | [ a , c ] ] &lt;~ [ a : 1 , b : 2 , c : 3 , d : 4 ] # =&gt; [ b: 2, a: 1, c: 3] [ a | [ b , c ( 25 ) ] ] &lt;~ %{ a : 1 , b : 2 } # =&gt; [a: 1, b: 2, c: 25] [ b | %{ c : %{ d } } ] &lt;~ [ a : 1 , b : 2 , c : %{ d : 5 } ] # =&gt; [ { :b, 2} | %{ a: 1, c: %{ d: 5}} Fetching non-optional values foo = Keyword . fetch! ( opts , :foo ) bar = Keyword . fetch! ( opts , :bar ) baz = Keyword . fetch! ( opts , :baz ) or foo = Map . fetch! ( opts , :foo ) bar = Map . fetch! ( opts , :bar ) baz = Map . fetch! ( opts , :baz ) can both be written simply as: [ foo , bar , baz ] &lt;~ opts # =&gt; fails if any of the three is not present in the opts Enforcing &quot;rigid&quot; (Elixir default) matching of the structures The rigid approach is useful to ensure an Elixir-like matching of the right side, and necessary if deconstructing standard Elixir tuples or non-keyword lists. ^ { a , b , c } &lt;~ { 1 , 2 , 3 } # ok ^ [ a , b , c ] &lt;~ [ 1 , 2 , 3 ] # ok ^ { b , a } &lt;~ { 1 , 2 , 3 } # error ^ [ a , b , c , d ] &lt;~ [ 1 , 2 , 3 ] # error ^ [ a , b , c : c ] &lt;~ [ 1 , 2 , 3 ] # error ^ %{ a } &lt;~ %{ a : 1 , b : 2 } # ok Destructuring from module (named) structures This is similar to destructuring from a map: [ hour , minute , second ] &lt;~ DateTime . utc_now ( ) # =&gt; [hour: 15, minute: 44, second: 14] or with the module key: [ __struct__ : module ] &lt;~ DateTime . utc_now ( ) # =&gt; [__struct__: DateTime] Nesting Any level of nesting is supported, and with the ^ operator toggling from loose to rigid and vice versa, any matching combination can be achieved. Ex from the extructure_test.ex : %{ a , b : b = ^ { c , d , ^ %{ e } } } &lt;~ [ a : 1 , b : { 3 , 4 , [ e : 5 ] } ] assert a == 1 assert b == { 3 , 4 , %{ e : 5 } } assert c == 3 assert d == 4 assert e == 5 Transforming the entire structure When in need to extract and transform an entire structure and not just some of its elements, all it takes is specifying an empty target structure (similar to Enum.into/2 , but consistent with the Extructure syntax, so that nesting is supported along with any other destructuring variables). Ex: [ a : a = [ ] ] &lt;~ [ a : [ b : 2 , c : 3 ] ] # a # =&gt; %{ b: 2, c: 3} or a = [ ] &lt;~ %{ b : 2 , c : 3 } # =&gt; [ b: 2, c: 3] An exceptional treatment of maps Unlike with lists and tuples, with maps the entire structure is transformed and associated with the corresponding variable even if there's only a subset of its variables getting destructured. Ex with destructuring into a map: [ a : a = %{ b } ] &lt;~ [ a : [ b : 2 , c : 3 ] ] # a =&gt; %{ b: 2, c: 3} # b =&gt; 2 Ex with destructuring into a list (same for tuples): [ a : a = [ b ] ] &lt;~ [ a : [ b : 2 , c : 3 ] ] # a =&gt; [ b: 2] # b =&gt; 2 String keys In addition to atom keys, Extructure supports destructuring from maps, key-value pair lists and key-value pair tuples with string keys. This feature is useful in such use cases as destructuring JSON properties or params in LiveView. All it takes is to prefix the intended part of the expression on the left with a @ character, e.g.: @ [ a ] &lt;~ %{ &quot;a&quot; =&gt; 1 } # a =&gt; 1 Just like ^ can be used in nested structures to toggle from loose to rigid mode and back, @ can be used to toggle from atom to string keys and back, e.g.: @ [ a , b : [ c , d : @ [ e ] ] ] &lt;~ %{ &quot;a&quot; =&gt; 1 , &quot;b&quot; =&gt; %{ &quot;c&quot; =&gt; 3 , &quot;d&quot; =&gt; [ e : 5 ] } } # a =&gt; 1 # c =&gt; 3 # e =&gt; 5 All matching restrictions apply same as with atom keys. Therefore, missing non-optional variables will result in failure while missing optional variables will not. @ [ a ] &lt;~ %{ &quot;b&quot; =&gt; 2 } # =&gt; MatchError @ [ a ( 1 ) ] &lt;~ %{ &quot;b&quot; =&gt; 2 } # a =&gt; 1 Key type toggling can be used in combination with mode toggling when needed, e.g.: @ ^ %{ a } &lt;~ %{ &quot;a&quot; =&gt; 1 } # =&gt; a = 1 @ ^ %{ a } &lt;~ [ { &quot;a&quot; , 1 } ] # =&gt; BadMapError","ref":"readme.html#sample-usage","title":"Extructure - Sample usage","type":"extras"},{"doc":"Optional variables and default values The original idea was to use the standard \\\\ operator to denote optional and default variables in lists, tuples, and maps, but, as shown below, the Elixir parser does not support this expression in maps. [ a , b \\\\ nil ] # ok { a , b \\\\ nil } # ok %{ a , b \\\\ nil } # syntax error So, decision was made that, until there's a progress with the Elixir parser, the underscore prefixed variable names will be used for optional variables defaulting to nil, and the function (macro) call syntax will be used for the optional variables defaulting to nil or any other value, e.g.: %{ _a } # optional variable, defaults to nil %{ a ( ) } # ditto %{ a ( 25 ) } # optional variable, defaults to 25 %{ _a ( 25 ) } # ditto The above syntax is used uniformly with all three types (maps, lists, tuples). The limitation that comes with this approach is that user-defined macro calls cannot be placed within the left-side expression. Should the Elixir core team decide to remove the parser restriction, a support for the standard Elixir optional variables (arguments) would be added and the present notation would be slowly phased out (leaving it to the compatibility mode).","ref":"readme.html#limitations","title":"Extructure - Limitations","type":"extras"},{"doc":"The source code formatting in this library diverges from the standard formatting practice based on using mix format in so much that there's a leading space character inserted before each initial argument / element with an intention to improve the code readability (subject to the author's personal perception). Another detail diverging from the standard Elixir formatting is that, where present, multi-line function signatures, and multi-line for , with , if , etc. statements will not have the do at the end of the last of the lines but, instead, indented in a new line, e.g.: with { _ , foo } &lt;- get_foo ( a , b , c ) , { _ , bar } &lt;- foo_to_bar ( foo ) do # logic else _ -&gt; x end The preferred width is 120 characters for the code and 80 characters for the docs.","ref":"readme.html#formatting","title":"Extructure - Formatting","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Enhancements Add support for string keys, e.g.: @ [ a , b , _c ] &lt;~ %{ &quot;a&quot; =&gt; 1 , &quot;b&quot; =&gt; 2 } # a =&gt; 1 # b =&gt; 2 # c =&gt; nil","ref":"changelog.html#v0-3-1-2023-05-07","title":"Changelog - v0.3.1 (2023-05-07)","type":"extras"},{"doc":"Enhancements Enable map destructuring as a whole even with a subset of keys specified When in loose mode we can typically use maps, lists and tuples interchangeably. However, to circumvent a problem that cannot be solved without completely redesigning the library (and making its execution much slower), an exception has been introduced in this version. The problem is related to the fact that the library is ultimately relying on pattern matching to associate values in a structure on the right (available only at runtime) with the variables on the left side (available for manipulation at compile time only). To illustrate the problem, consider what the two statements below translate into: [ a : a = %{ b } ] &lt;~ term # translates into: [ a : a = %{ b : b } ] = adjusted_term while [ a : a = [ b ] ] &lt;~ term # translates into: [ a : a = [ b : b ] ] = adjusted_term The former automatically associates an entire structure in term with the variable a , not just its subset with the variable b only, while the latter associates the keyword list with just the b key (note: the Extructure logic drops all key-value pairs except for the b in runtime so the pattern matching does not fail). In previous versions the logic used to drop all undeclared keys from the maps as well in order to enforce a behavioral similarity with lists and tuples, but this approach came with a tradeoff - not being able to destructure both the structure as a whole and some of its elements in a single statement. This is why this restriction has been removed in the v0.3.0. Bug fixes Destructure module structure as a whole as the module structure (not as a plain map).","ref":"changelog.html#v0-3-0-2023-04-29","title":"Changelog - v0.3.0 (2023-04-29)","type":"extras"},{"doc":"Bug fixes Fix the flaw with a list or a tuple not failing when non optional variables are missing. The following used to pass but now fails as supposed to: [ a ] &lt;~ [ b : 1 ] # =&gt; MatchError Fix the flaw with rigid list and tuple not supporting unnamed underscore variables. [ x : ^ { a , _ , _ , d } ] &lt;~ %{ x : { 1 , 2 , 3 , 4 } } # =&gt; [x: {1, 2, 3, 4}] # a # =&gt; 1 # d # =&gt; 4","ref":"changelog.html#v0-2-2-2023-04-16","title":"Changelog - v0.2.2 (2023-04-16)","type":"extras"},{"doc":"Enhancements Support transforming entire structures on the right side by specifying an empty map, tuple or list, e.g.: [ a : a = %{ } ] &lt;~ [ a : [ b : 2 , c : 3 ] ] # a # =&gt; %{b: 2, c: 3} Support destructuring from module (named) structures as if plain maps, e.g: [ a , b ] &lt;~ % Foo { a : 1 , b : 2 } # =&gt; [a: 1, b: 2] Bug fixes Make maps on the left dictate the keys to merge in the same way lists and tuples do.","ref":"changelog.html#v0-2-1-2023-02-26","title":"Changelog - v0.2.1 (2023-02-26)","type":"extras"},{"doc":"Breaking Changes Make rigid lists match identically as in standard Elixir matching. It is necessary to depart from any looseness in rigid lists as otherwise, a standard pattern matching based destructuring is not possible (always converting the variables to variable pairs). Ex: ^ [ a , b ] &lt;~ [ 1 , 2 ] # =&gt; [ 1, 2] ^ [ a , b ] &lt;~ [ 1 , 2 , 3 ] # =&gt; MatchError Deprecations Warn when optional variables used with rigid lists or tuples for it no longer makes no sense.","ref":"changelog.html#v0-2-0-2023-02-19","title":"Changelog - v0.2.0 (2023-02-19)","type":"extras"},{"doc":"Enhancements Added support for head | tail destructure e.g. [ b | rest ] &lt;~ [ a : 1 , b : 2 , c : 3 ] # =&gt; [ b: 2, a: 1, c: 3]","ref":"changelog.html#v0-1-1-2023-02-11","title":"Changelog - v0.1.1 (2023-02-11)","type":"extras"},{"doc":"Initial implementation","ref":"changelog.html#v0-1-0-2023-02-05","title":"Changelog - v0.1.0 (2023-02-05)","type":"extras"}]